import sys
import unittest

import pytest

import semeio.hook_implementations.jobs
from ert_shared.plugins.plugin_manager import ErtPluginManager


@pytest.mark.skipif(sys.version_info.major < 3, reason="requires python3")
class Test(unittest.TestCase):
    def test_hook_implementations(self):
        pm = ErtPluginManager(plugins=[semeio.hook_implementations.jobs])
        self.assertDictEqual({}, pm.get_installable_jobs())
        self.assertDictEqual({}, pm.get_installable_workflow_jobs())



import os
import shutil

import configsuite
import numpy as np
import pytest
import yaml
from res.enkf import EnKFMain, ResConfig
from ert_shared.libres_facade import LibresFacade
from semeio.jobs.correlated_observations_scaling import (
    job as scaling_job,
    job_config,
    validator,
)
from tests.jobs.correlated_observations_scaling.conftest import TEST_DATA_DIR


def assert_obs_vector(vector, val_1, index_list=None, val_2=None):
    if index_list is None:
        index_list = []
    for node in vector:
        for index in range(len(node)):
            if index in index_list:
                assert node.getStdScaling(index) == val_2
            else:
                assert node.getStdScaling(index) == val_1


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.fixture()
def setup_tmpdir(tmpdir):
    cwd = os.getcwd()
    tmpdir.chdir()
    yield
    os.chdir(cwd)


def get_config(index_list_calc=None, index_list_update=None):
    schema = job_config.build_schema()
    default_values = job_config.get_default_values()
    config = {
        "UPDATE_KEYS": {"keys": [{"key": "POLY_OBS"}]},
        "CALCULATE_KEYS": {"keys": [{"key": "POLY_OBS"}]},
    }

    if index_list_update:
        config["UPDATE_KEYS"]["keys"][0].update({"index": index_list_update})
    if index_list_calc:
        config["CALCULATE_KEYS"]["keys"][0].update({"index": index_list_calc})

    return configsuite.ConfigSuite(config, schema, layers=(default_values,)).snapshot


def get_job_dir():
    from semeio import jobs

    head, _ = os.path.split(os.path.abspath(jobs.__file__))
    return os.path.join(head, "configs")


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_ert")
def test_old_enkf_scaling_job(setup_ert):

    res_config = setup_ert
    ert = EnKFMain(res_config)

    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0)

    job = ert.getWorkflowList().getJob("STD_SCALE_CORRELATED_OBS")
    job.run(ert, ["WPR_DIFF_1"])

    assert_obs_vector(obs_vector, np.sqrt(4.0 / 2.0))


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_ert")
def test_installed_python_version_of_enkf_scaling_job(setup_ert, monkeypatch):
    res_config = setup_ert
    ert = EnKFMain(res_config)

    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0)

    job_config = {"CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]}}

    with open("job_config.yml", "w") as fout:
        yaml.dump(job_config, fout)

    ert.getWorkflowList().addJob(
        "CORRELATE_OBSERVATIONS_SCALING",
        os.path.join(get_job_dir(), "CORRELATED_OBSERVATIONS_SCALING"),
    )

    job = ert.getWorkflowList().getJob("CORRELATE_OBSERVATIONS_SCALING")
    job.run(ert, ["job_config.yml"])

    assert_obs_vector(obs_vector, np.sqrt(4.0 / 2.0))

    job_config["CALCULATE_KEYS"]["keys"][0].update({"index": [400, 800, 1200]})
    with open("job_config.yml", "w") as fout:
        yaml.dump(job_config, fout)
    job.run(ert, ["job_config.yml"])

    assert_obs_vector(
        obs_vector, np.sqrt(4.0 / 2.0), index_list=[0, 1, 2], val_2=np.sqrt(3.0 / 2.0),
    )


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_compare_different_jobs():

    arguments = {"CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]}}

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)
    facade = LibresFacade(ert)
    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0)

    job = ert.getWorkflowList().getJob("STD_SCALE_CORRELATED_OBS")
    job.run(ert, ["WPR_DIFF_1"])

    # Result of old job:
    assert_obs_vector(obs_vector, np.sqrt(4 / 2))

    scaling_job.scaling_job(facade, arguments)

    # Result of new job with no sub-indexing:
    assert_obs_vector(obs_vector, np.sqrt(4 / 2))


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_gen_data():

    arguments = {
        "CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WPR_DIFF_1", "index": [400, 800]}]},
    }

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)
    facade = LibresFacade(ert)

    scaling_job.scaling_job(facade, arguments)

    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0, [0, 1], np.sqrt(4 / 2))

    arguments["CALCULATE_KEYS"]["keys"][0].update({"index": [400, 800, 1200]})
    scaling_job.scaling_job(facade, arguments)
    assert_obs_vector(
        obs_vector, 1.0, [0, 1], np.sqrt(3.0 / 2.0),
    )


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_summary_data_calc():

    arguments = {
        "CALCULATE_KEYS": {"keys": [{"key": "WOPR_OP1_108"}, {"key": "WOPR_OP1_144"}]}
    }

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)
    facade = LibresFacade(ert)
    obs = ert.getObservations()

    obs_vector = obs["WOPR_OP1_108"]

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0

    scaling_job.scaling_job(facade, arguments)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == np.sqrt((2.0 * 6.0) / 2.0)

    arguments["CALCULATE_KEYS"]["keys"][0].update({"index": [1, 2, 3]})
    arguments["CALCULATE_KEYS"]["keys"][1].update({"index": [1, 2, 3]})

    with pytest.raises(ValueError):  # Will give an empty data set
        scaling_job.scaling_job(facade, arguments)

    arguments["CALCULATE_KEYS"]["keys"][0].update({"index": [8, 35, 71]})
    arguments["CALCULATE_KEYS"]["keys"][1].update({"index": [8, 35, 71]})
    scaling_job.scaling_job(facade, arguments)

    for index, node in enumerate(obs_vector):
        if index in arguments["CALCULATE_KEYS"]["keys"][0]["index"]:
            assert node.getStdScaling(index) == np.sqrt((2.0 * 6.0) / 1.0)
        else:
            assert node.getStdScaling(index) == np.sqrt((2.0 * 6.0) / 2.0)


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.equinor_test
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_summary_data_update():
    arguments = {
        "CALCULATE_KEYS": {"keys": [{"key": "WWCT:OP_1"}, {"key": "WWCT:OP_2"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WWCT:OP_2", "index": [1, 2, 3, 4, 5]}]},
    }

    test_data_dir = os.path.join(TEST_DATA_DIR, "Equinor", "config", "obs_testing")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("config")
    ert = EnKFMain(res_config)
    facade = LibresFacade(ert)
    obs = ert.getObservations()
    obs_vector = obs["WWCT:OP_2"]

    scaling_job.scaling_job(facade, arguments)

    for index, node in enumerate(obs_vector):
        if index in arguments["UPDATE_KEYS"]["keys"][0]["index"]:
            assert node.getStdScaling(index) == np.sqrt(61.0 * 2.0)
        else:
            assert node.getStdScaling(index) == 1.0

    obs_vector = obs["WWCT:OP_1"]

    scaling_job.scaling_job(facade, arguments)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.equinor_test
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_block_data_calc():
    arguments = {"CALCULATE_KEYS": {"keys": [{"key": "RFT3"}]}}

    test_data_dir = os.path.join(TEST_DATA_DIR, "Equinor", "config", "with_RFT")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("config")
    ert = EnKFMain(res_config)
    facade = LibresFacade(ert)
    obs = ert.getObservations()

    obs_vector = obs["RFT3"]

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0

    scaling_job.scaling_job(facade, arguments)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 2.0


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.equinor_test
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_block_and_summary_data_calc():
    arguments = {"CALCULATE_KEYS": {"keys": [{"key": "FOPT"}, {"key": "RFT3"}]}}

    test_data_dir = os.path.join(TEST_DATA_DIR, "Equinor", "config", "with_RFT")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("config")
    ert = EnKFMain(res_config)
    facade = LibresFacade(ert)
    obs = ert.getObservations()

    obs_vector = obs["RFT3"]

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0

    scaling_job.scaling_job(facade, arguments)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == np.sqrt(64)


def test_valid_configuration():

    valid_config_data = {
        "CALCULATE_KEYS": {"keys": [{"key": "POLY_OBS"}]},
        "UPDATE_KEYS": {"keys": [{"key": "POLY_OBS"}]},
    }

    schema = job_config.build_schema()
    config = configsuite.ConfigSuite(valid_config_data, schema)

    assert validator.valid_configuration(config)



import numpy as np

from copy import deepcopy


class DataMatrix(object):
    def __init__(self, input_data):
        """
            Takes input data in the form of a Pandas multi index dataframe with observations,
        standard deviation and simulated values. Assumes observations are
        prepended with _OBS and standard deviation with _STD.
        """
        self.data = input_data
        if input_data.shape[1] == 0:
            raise ValueError("Empty dataset, all data has been filtered out")

    def get_data_matrix(self):
        """
        Extracts data from a dataframe and returns a matrix
        """
        return self.data[~self.data.index.isin(["OBS", "STD"])].values

    def std_normalization(self, inplace=False):
        """
        Duplicates the behavior of obs_data_scale, and scales the simulated data
        by 1 / (observation standard deviation), per observation key, i.e. each
        simulation data point is scaled by its corresponding std deviation
        from observations.
        """
        output_data = deepcopy(self.data)
        data_matrix = self._get_data()
        std_vector = self.data.loc["STD"]
        output_data[~output_data.index.isin(["OBS", "STD"])] = data_matrix * (
            1.0 / std_vector
        )

        if inplace:
            self.data = output_data
        return output_data

    def get_scaling_factor(self, events):
        """
        Performs PCA, calculates the number of primary components based on
        a threshold and returns a scaling factor based on the number of
        primary components and the number of observations.
        """
        data_matrix = self.get_data_matrix()
        nr_components = self._get_nr_primary_components(
            data_matrix, threshold=events.threshold
        )
        scaling_factor = self._calculate_scaling_factor(
            data_matrix.shape[1], nr_components
        )

        print("Scaling factor calculated from {}".format(events.keys))
        return scaling_factor

    def _get_data(self):
        return self.data[~self.data.index.isin(["OBS", "STD"])]

    @staticmethod
    def _get_nr_primary_components(data_matrix, threshold):
        """
        Takes a matrix, does PCA and calculates the cumulative variance ratio
        and returns an int which is the number of primary components where
        the cumulative variance is smaller than user set threshold. Notice the
        way of calculating number of primary components. This is done to
        replicate existing behavior:

        int num_significant  = 0;
        {
        double running_sigma2  = 0;
        for (int i=0; i < num_singular_values; i++) {
          if (running_sigma2 / total_sigma2 < truncation) {
             num_significant++;
             running_sigma2 += sig0[i] * sig0[i];
          } else
             break;
        }
        """
        _, s, _ = np.linalg.svd(data_matrix.astype(np.float), full_matrices=False)
        variance_ratio = np.cumsum(s ** 2) / np.sum(s ** 2)
        return len([1 for i in variance_ratio[:-1] if i < threshold]) + 1

    @staticmethod
    def _calculate_scaling_factor(nr_observations, nr_components):
        """
        Calculates a observation scaling factor which is:
            sqrt(nr_obs / pc)
        where:
            nr_obs is the number of observations
            pc is the number of primary components from PCA analysis
                below a user threshold
        """
        print(
            "Calculation scaling factor, nr of primary components: {:d}, number of observations: {:d}".format(
                nr_components, nr_observations
            )
        )
        return np.sqrt(nr_observations / float(nr_components))



# -*- coding: utf-8 -*-
import itertools

from ert_data.measured import MeasuredData
from scipy.cluster.hierarchy import linkage, fcluster
from semeio.jobs.correlated_observations_scaling.job import scaling_job


def spearman_job(facade, threshold, dry_run):

    observation_keys = [
        facade.get_observation_key(nr) for nr, _ in enumerate(facade.get_observations())
    ]

    _spearman_correlation(facade, observation_keys, threshold, dry_run)


def _spearman_correlation(facade, obs_keys, threshold, dry_run):
    """
    Collects data, performs scaling and applies scaling, assumes validated input.
    """
    measured_data = MeasuredData(facade, obs_keys)
    measured_data.remove_failed_realizations()
    measured_data.remove_inactive_observations()
    measured_data.filter_ensemble_std(1.0e-6)

    simulated_data = measured_data.get_simulated_data()

    correlation_matrix = _calculate_correlation_matrix(simulated_data)

    clusters = _cluster_analysis(correlation_matrix, threshold)

    columns = correlation_matrix.columns

    # Here the clusters are joined with the key and data index
    # to group the observations, the column level values are the column
    # headers, where key_index is the observation key and data_index
    # is a range.
    data = list(
        zip(
            clusters,
            columns.get_level_values(0),
            columns.get_level_values("data_index"),
        )
    )

    clustered_data = _cluster_data(data)

    job_configs = _config_creation(clustered_data)

    _output_clusters(clustered_data)

    if not dry_run:
        _run_scaling(facade, job_configs)


def _output_clusters(clustered_data):
    for cluster, val in clustered_data.items():
        print("Cluster nr: {}, clustered data: {}".format(cluster, val))


def _run_scaling(facade, job_configs):
    for job in job_configs:
        scaling_job(facade, job)


def _cluster_data(data):
    groups = {}
    for (nr, key), cluster_group in itertools.groupby(
        sorted(data), key=lambda x: (x[0], x[1])
    ):
        if nr not in groups:
            groups[nr] = {}
        groups[nr].update({key: [index for _, _, index in cluster_group]})
    return groups


def _config_creation(clusters):
    config = []
    for cluster_nr, cluster in clusters.items():
        config.append(
            {
                "CALCULATE_KEYS": {
                    "keys": [{"key": key, "index": val} for key, val in cluster.items()]
                }
            }
        )
    return config


def _calculate_correlation_matrix(data):
    # Spearman correlation is quite slow, but will be improved in a future version
    # of pandas (https://github.com/pandas-dev/pandas/pull/28151), for now this is
    # equivalent:
    return data.rank().corr(method="pearson")


def _cluster_analysis(correlation_matrix, threshold):
    a = linkage(correlation_matrix, "single")
    return fcluster(a, threshold)



def valid_job(user_config, obs_list, obs_with_data):
    """
    Validates the job, assumes that the configuration is valid
    """

    calc_keys = [entry.key for entry in user_config.snapshot.CALCULATE_KEYS.keys]
    application_keys = [entry.key for entry in user_config.snapshot.UPDATE_KEYS.keys]

    error_messages = []

    error_messages.extend(_is_subset(calc_keys, application_keys))
    error_messages.extend(_has_keys(obs_list, calc_keys, "Key: {} has no observations"))
    error_messages.extend(_has_keys(obs_with_data, calc_keys, "Key: {} has no data"))

    for error in error_messages:
        print(error)
    return len(error_messages) == 0


def _has_keys(main_list, sub_list, error_msg):
    """
    Checks that all sub_list are present and returns a list of error messages
    """
    return [error_msg.format(key) for key in sub_list if key not in main_list]


def _is_subset(main_list, sub_list):
    """
    Checks if all the keys in sub_list are present in main_list and returns list of error
    messages
    """
    error_msg = "Update key: {} missing from calculate keys: {}"
    missing_keys = set(sub_list).difference(set(main_list))
    return [error_msg.format(missing_key, main_list) for missing_key in missing_keys]


def valid_configuration(user_config):
    """
    Validates the configuration
    """
    for error in user_config.errors:
        print(error)
    return user_config.valid



import os
import shutil
import sys

import pandas as pd
import pytest
from res.enkf import EnKFMain, ResConfig

from ert_shared.libres_facade import LibresFacade
from semeio.jobs.spearman_correlation_job import job as spearman
from tests.jobs.correlated_observations_scaling.conftest import TEST_DATA_DIR

if sys.version_info >= (3, 3):
    from unittest.mock import Mock
else:
    from mock import Mock


def test_spearman_correlation(monkeypatch):
    df = pd.DataFrame(data=[[7, 8, 9], [10, 11, 12]], index=[1, 2], columns=[0, 1, 2])
    tuples = list(zip(*[df.columns.to_list(), df.columns.to_list()]))
    df.columns = pd.MultiIndex.from_tuples(tuples, names=["key_index", "data_index"])

    facade = Mock()
    mock_data = Mock()
    mock_data.get_simulated_data.return_value = df
    measured_data = Mock(return_value=mock_data)
    scal_job = Mock()
    monkeypatch.setattr(spearman, "scaling_job", scal_job)
    monkeypatch.setattr(spearman, "MeasuredData", measured_data)
    spearman._spearman_correlation(facade, ["A_KEY"], 0.1, False)

    assert measured_data.called_once_with(facade, ["A_KEY"])
    assert scal_job.called_once()


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_gen_data(monkeypatch):
    scal_job = Mock()
    monkeypatch.setattr(spearman, "scaling_job", scal_job)

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)
    facade = LibresFacade(ert)

    spearman.spearman_job(facade, 1.0, False)

    assert scal_job.call_count == 71



import argparse

from ert_shared.libres_facade import LibresFacade
from res.enkf import ErtScript

from semeio.jobs.spearman_correlation_job.job import spearman_job


class SpearmanCorrelationJob(ErtScript):
    def run(self, *args):
        facade = LibresFacade(self.ert())

        parser = spearman_job_parser()
        args = parser.parse_args(args)

        spearman_job(facade, args.threshold, args.dry_run)


def spearman_job_parser():
    description = """
    A module that calculates the Spearman correlation in simulated data
    and clusters
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-t",
        "--threshold",
        required=False,
        default=1.15,
        type=float,
        help="""
        Forms flat clusters so that the original
        observations in each flat cluster have no greater a
        cophenetic distance than `t`.
        """,
    )
    parser.add_argument(
        "--output-file",
        required=False,
        type=str,
        help="Name of the outputfile. The format will be yaml.",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        required=False,
        help="Dry run, no scaling will be performed",
        action="store_true",
    )
    return parser



# -*- coding: utf-8 -*-
import sys

import pytest
from semeio.jobs.spearman_correlation_job import job as spearman

if sys.version_info >= (3, 3):
    from unittest.mock import Mock
else:
    from mock import Mock


@pytest.mark.parametrize(
    "test_input",
    [
        ([{"CALCULATE_KEYS": {"keys": [{"key": "KEY_1", "index": [1]}]}}]),
        ([{"CALCULATE_KEYS": {"keys": [{"key": "KEY_1", "index": [1, 2]}]}}]),
        (
            [
                {
                    "CALCULATE_KEYS": {
                        "keys": [
                            {"key": "KEY_1", "index": [1, 2]},
                            {"key": "KEY_2", "index": [1, 2]},
                        ]
                    }
                }
            ]
        ),
        (
            [
                {"CALCULATE_KEYS": {"keys": [{"key": "KEY_1", "index": [1, 2]}]}},
                {"CALCULATE_KEYS": {"keys": [{"key": "KEY_2", "index": [5, 6]}]}},
            ]
        ),
    ],
)
def test_run_scaling(test_input, monkeypatch):
    facade = Mock()
    scal_job = Mock()
    monkeypatch.setattr(spearman, "scaling_job", scal_job)
    spearman._run_scaling(facade, test_input)
    for input_config in test_input:
        assert scal_job.any_call(input_config)
    assert scal_job.call_count == len(test_input)


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        ([(1, "KEY_1", 1)], {1: {"KEY_1": [1]}}),
        ([(1, "KEY_1", 1), (1, "KEY_1", 2)], {1: {"KEY_1": [1, 2]}}),
        ([(1, "KEY_1", 1), (2, "KEY_2", 2)], {1: {"KEY_1": [1]}, 2: {"KEY_2": [2]}}),
    ],
)
def test_make_clusters(test_input, expected_result):
    result = spearman._cluster_data(test_input)
    assert result == expected_result


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (
            {1: {"KEY_1": [1]}},
            [{"CALCULATE_KEYS": {"keys": [{"key": "KEY_1", "index": [1]}]}}],
        )
    ],
)
def test_config_creation(test_input, expected_result):
    result = spearman._config_creation(test_input)
    assert result == expected_result



import sys
import pytest

from semeio.jobs.correlated_observations_scaling import validator

if sys.version_info >= (3, 3):
    from unittest.mock import Mock
else:
    from mock import Mock


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (["POLY_OBS"], 0),
        (["NOT_A_KEY"], 1),
        (["POLY_OBS", "NOT_A_KEY"], 1),
        (["POLY_NOT", "NOT_A_KEY"], 2),
    ],
)
def test_has_keys(test_input, expected_result):
    obs = ["POLY_OBS"]
    msg = "fail_message"
    assert len(validator._has_keys(obs, test_input, msg)) == expected_result


@pytest.mark.parametrize(
    "input_list,result_list", [(["a"], []), (["a", "c"], []), (["a", "b", "c"], [])]
)
def test_is_subset_valid(input_list, result_list):
    example_list = ["a", "b", "c"]

    assert validator._is_subset(example_list, input_list) == result_list


@pytest.mark.parametrize(
    "input_list,list_length", [(["d"], 1), (["d", "e"], 2), (["a", "b", "d"], 1)]
)
def test_is_subset_invalid(input_list, list_length):
    example_list = ["a", "b", "c"]

    assert len(validator._is_subset(example_list, input_list)) == list_length


@pytest.mark.parametrize(
    "calc_key,app_key,obs_list,data_list,expected_result",
    [
        ("KEY_1", "KEY_1", ["KEY_1"], ["KEY_1"], True),
        ("KEY_1", "KEY_1", ["KEY_1", "KEY_2"], ["KEY_1"], True),
        ("KEY_1", "KEY_1", ["KEY_1", "KEY_2"], ["KEY_2"], False),
        ("not_in_list", "KEY_1", ["KEY_1"], ["KEY_1"], False),
        ("KEY_1", "not_in_list", ["KEY_1"], ["KEY_1"], False),
        ("KEY_1", "KEY_1", [], ["KEY_1"], False),
        ("KEY_1", "KEY_1", ["KEY_1"], [], False),
    ],
)
def test_valid_job(calc_key, app_key, obs_list, data_list, expected_result):

    mock_entry_1 = Mock()
    mock_entry_1.key = calc_key

    mock_entry_2 = Mock()
    mock_entry_2.key = app_key

    user_config = Mock()
    user_config.snapshot.CALCULATE_KEYS.keys = [mock_entry_1]
    user_config.snapshot.UPDATE_KEYS.keys = [mock_entry_2]

    assert validator.valid_job(user_config, obs_list, data_list) is expected_result



import collections

import yaml
from ert_shared.libres_facade import LibresFacade
from res.enkf import ErtScript

from semeio.jobs.correlated_observations_scaling.job import scaling_job


class CorrelatedObservationsScalingJob(ErtScript):
    def run(self, job_config_file):
        facade = LibresFacade(self.ert())
        user_config = load_yaml(job_config_file)
        user_config = _insert_default_group(user_config)
        for job_config in user_config:
            scaling_job(facade, job_config)


def load_yaml(f_name):
    with open(f_name, "r") as fin:
        config = yaml.safe_load(fin)
    return config


def _insert_default_group(value):
    if isinstance(value, collections.Mapping):
        return [value]
    return value



# -*- coding: utf-8 -*-
import fnmatch

import configsuite

from copy import deepcopy
from collections import namedtuple

from semeio.jobs.correlated_observations_scaling import job_config
from semeio.jobs.correlated_observations_scaling.scaled_matrix import DataMatrix
from ert_data.measured import MeasuredData
from semeio.jobs.correlated_observations_scaling.validator import (
    valid_configuration,
    valid_job,
)
from res.enkf import LocalObsdata, ActiveList

from ecl.util.util import BoolVector
from res.enkf import RealizationStateEnum


def scaling_job(facade, user_config_dict):
    """
    Takes an instance of EnkFMain and a user config dict, will do some pre-processing on
    the user config dict, set up a ConfigSuite instance and validate the job before control
    is passed to the main job.
    """

    observation_keys = [
        facade.get_observation_key(nr) for nr, _ in enumerate(facade.get_observations())
    ]
    obs_with_data = keys_with_data(
        facade.get_observations(),
        observation_keys,
        facade.get_ensemble_size(),
        facade.get_current_fs(),
    )
    config_dict = _find_and_expand_wildcards(observation_keys, user_config_dict)

    config = setup_configuration(config_dict, job_config.build_schema())

    if not valid_configuration(config):
        raise ValueError("Invalid configuration")
    if not valid_job(config, observation_keys, obs_with_data):
        raise ValueError("Invalid job")
    _observation_scaling(facade, config.snapshot)


def _observation_scaling(facade, config):
    """
    Collects data, performs scaling and applies scaling, assumes validated input.
    """
    calculate_keys = [event.key for event in config.CALCULATE_KEYS.keys]
    index_lists = [event.index for event in config.CALCULATE_KEYS.keys]
    measured_data = MeasuredData(facade, calculate_keys, index_lists)
    measured_data.remove_failed_realizations()
    measured_data.remove_inactive_observations()
    measured_data.filter_ensemble_mean_obs(config.CALCULATE_KEYS.alpha)
    measured_data.filter_ensemble_std(config.CALCULATE_KEYS.std_cutoff)

    matrix = DataMatrix(measured_data.data)
    matrix.std_normalization(inplace=True)

    scale_factor = matrix.get_scaling_factor(config.CALCULATE_KEYS)

    update_data = _create_active_lists(
        facade.get_observations(), config.UPDATE_KEYS.keys
    )

    _update_scaling(facade.get_observations(), scale_factor, update_data)


def _wildcard_to_dict_list(matching_keys, entry):
    """
    One of either:
    entry = {"key": "WOPT*", "index": [1,2,3]}
    entry = {"key": "WOPT*"}
    """
    if "index" in entry:
        return [{"key": key, "index": entry["index"]} for key in matching_keys]
    else:
        return [{"key": key} for key in matching_keys]


def _expand_wildcard(obs_list, wildcard_key, entry):
    """
    Expands a wildcard
    """
    matching_keys = fnmatch.filter(obs_list, wildcard_key["key"])
    return _wildcard_to_dict_list(matching_keys, entry)


def _find_and_expand_wildcards(obs_list, user_dict):
    """
    Loops through the user input and identifies wildcards in observation
    names and expands them.
    """
    new_dict = deepcopy(user_dict)
    for main_key, value in user_dict.items():
        new_entries = []
        if main_key in ("UPDATE_KEYS", "CALCULATE_KEYS"):
            for val in value["keys"]:
                if "*" in val["key"]:
                    new_entries.extend(_expand_wildcard(obs_list, val, val["key"]))
                else:
                    new_entries.append(val)
            new_dict[main_key]["keys"] = new_entries

    return new_dict


def setup_configuration(input_data, schema):
    """
    Creates a ConfigSuite instance and inserts default values
    """
    default_layer = job_config.get_default_values()
    config = configsuite.ConfigSuite(input_data, schema, layers=(default_layer,))
    return config


def _create_active_lists(enkf_observations, events):
    """
    Will add observation vectors to observation data. Returns
    a list of tuples mirroring the user config but also containing
    the active list where the scaling factor will be applied.
    """
    new_events = []
    observation_data = LocalObsdata("some_name", enkf_observations)
    for event in events:
        observation_data.addObsVector(enkf_observations[event.key])

        obs_index = _data_index_to_obs_index(enkf_observations, event.key, event.index)
        new_active_list = _get_active_list(observation_data, event.key, obs_index)

        new_events.append(_make_tuple(event.key, event.index, new_active_list))

    return new_events


def _get_active_list(observation_data, key, index_list):
    """
    If the user doesn't supply an index list, the existing active
    list from the observation is used, otherwise an active list is
    created from the index list.
    """
    if index_list is not None:
        return _active_list_from_index_list(index_list)
    else:
        return observation_data.copy_active_list(key).setParent()


def _make_tuple(
    key,
    index,
    active_list,
    new_event=namedtuple("named_dict", ["key", "index", "active_list"]),
):
    return new_event(key, index, active_list)


def _update_scaling(obs, scale_factor, events):
    """
    Applies the scaling factor to the user specified index, SUMMARY_OBS needs to be treated differently
    as it only has one data point per node, compared with other observation types which have multiple
    data points per node.
    """
    for event in events:
        obs_vector = obs[event.key]
        for index, obs_node in enumerate(obs_vector):
            if obs_vector.getImplementationType().name == "SUMMARY_OBS":
                index_list = (
                    event.index if event.index is not None else range(len(obs_vector))
                )
                if index in index_list:
                    obs_node.set_std_scaling(scale_factor)
            elif obs_vector.getImplementationType().name != "SUMMARY_OBS":
                obs_node.updateStdScaling(scale_factor, event.active_list)
    print(
        "Keys: {} scaled with scaling factor: {}".format(
            [event.key for event in events], scale_factor
        )
    )


def _active_list_from_index_list(index_list):
    """
    Creates an ActiveList from a list of indexes
    :param index_list: list of index
    :type index_list:  list
    :return: Active list, a c-object with mode (ALL-ACTIVE, PARTIALLY-ACTIVE, INACTIVE) and list of indices
    :rtype: active_list
    """
    active_list = ActiveList()
    [active_list.addActiveIndex(index) for index in index_list]
    return active_list


def _set_active_lists(observation_data, key_list, active_lists):
    """
    Will make a backup of the existing active list on the observation node
    before setting the user supplied index list.
    """
    exisiting_active_lists = []
    for key, active_list in zip(key_list, active_lists):
        exisiting_active_lists.append(observation_data.copy_active_list(key))
        observation_data.setActiveList(key, active_list)
    return observation_data, exisiting_active_lists


def keys_with_data(observations, keys, ensamble_size, storage):
    """
    Checks that all keys have data and returns a list of error messages
    """
    active_realizations = storage.realizationList(RealizationStateEnum.STATE_HAS_DATA)

    if len(active_realizations) == 0:
        return []

    active_mask = BoolVector.createFromList(ensamble_size, active_realizations)
    return [key for key in keys if observations[key].hasData(active_mask, storage)]


def _data_index_to_obs_index(obs, obs_key, data_index_list):
    if obs[obs_key].getImplementationType().name != "GEN_OBS":
        return data_index_list
    elif data_index_list is None:
        return data_index_list

    for timestep in obs[obs_key].getStepList().asList():
        node = obs[obs_key].getNode(timestep)
        index_map = {node.getIndex(nr): nr for nr in range(len(node))}
    return [index_map[index] for index in data_index_list]



import os
import shutil
from collections import namedtuple
from copy import deepcopy

import configsuite
import numpy as np
import pandas as pd
import pytest
from res.enkf import EnKFMain, ResConfig

from ert_data import measured
from semeio.jobs.correlated_observations_scaling import (
    job,
    job_config,
    scaled_matrix,
    validator,
)
from tests.jobs.correlated_observations_scaling.conftest import TEST_DATA_DIR


@pytest.fixture()
def setup_tmpdir(tmpdir):
    cwd = os.getcwd()
    tmpdir.chdir()
    yield
    os.chdir(cwd)


def get_config(index_list_calc=None, index_list_update=None):
    schema = job_config.build_schema()
    default_values = job_config.get_default_values()
    config = {
        "UPDATE_KEYS": {"keys": [{"key": "POLY_OBS"}]},
        "CALCULATE_KEYS": {"keys": [{"key": "POLY_OBS"}]},
    }

    if index_list_update:
        config["UPDATE_KEYS"]["keys"][0].update({"index": index_list_update})
    if index_list_calc:
        config["CALCULATE_KEYS"]["keys"][0].update({"index": index_list_calc})

    return configsuite.ConfigSuite(config, schema, layers=(default_values,)).snapshot


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_ert")
def test_create_observation_vectors(setup_ert):

    valid_config_data = {
        "CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]},
    }
    config = configsuite.ConfigSuite(valid_config_data, job_config.build_schema())

    res_config = setup_ert
    ert = EnKFMain(res_config)
    obs = ert.getObservations()

    new_events = job._create_active_lists(obs, config.snapshot.UPDATE_KEYS.keys)

    keys = [event.key for event in new_events]

    assert "WPR_DIFF_1" in keys
    assert "SNAKE_OIL_WPR_DIFF" not in keys


@pytest.mark.parametrize(
    "valid_input",
    [
        "0,1,2-5",
        [0, 1, "2-5"],
        [0, 1, 2, 3, 4, 5],
        [0, 1, "2-3", "4-5"],
        "0-5",
        "0-1,2,3-5",
        ["0,1,2-5"],
    ],
)
def test_to_int_list(valid_input):
    expected_result = list(range(6))
    assert job_config._to_int_list(valid_input) == expected_result


@pytest.mark.parametrize(
    "test_input,expected_result", [(-1, False), (0, True), (1, True)]
)
def test_min_value(test_input, expected_result):
    assert job_config._min_value(test_input).__bool__() == expected_result


def test_expand_input_no_modification():

    expected_result = {
        "UPDATE_KEYS": {"keys": [{"key": "key_4"}, {"key": "key_5"}, {"key": "key_6"}]},
        "CALCULATE_KEYS": {
            "keys": [{"key": "key_1"}, {"key": "key_2"}, {"key": "key_3"}],
            "threshold": 1.0,
        },
    }

    assert job_config._expand_input(deepcopy(expected_result)) == expected_result


def test_expand_input_modification():
    keys = [{"key": "key_1"}, {"key": "key_2"}, {"key": "key_3"}]
    test_input = {"CALCULATE_KEYS": {"keys": keys, "threshold": 1.0}}

    expected_results = deepcopy(test_input)
    expected_results["UPDATE_KEYS"] = {"keys": keys}

    assert job_config._expand_input(test_input) == expected_results


def test_config_setup():

    valid_config_data = {
        "CALCULATE_KEYS": {"keys": [{"key": "first_key"}, {"key": "second_key"}]}
    }

    schema = job_config.build_schema()
    config = configsuite.ConfigSuite(valid_config_data, schema)
    assert config.valid

    valid_config_data = {
        "CALCULATE_KEYS": {"keys": [{"key": "first_key"}, {"key": "second_key"}]},
        "UPDATE_KEYS": {"keys": [{"index": [1, 2, 3], "key": "first_key"}]},
    }

    schema = job_config.build_schema()
    config = configsuite.ConfigSuite(valid_config_data, schema)
    assert config.valid

    invalid_too_short_index_list = {
        "UPDATE_KEYS": {"keys": [{"index": "1", "key": ["a_key"]}]}
    }

    config = configsuite.ConfigSuite(invalid_too_short_index_list, schema)
    assert not config.valid

    invalid_missing_required_keyword = {
        "CALCULATE_KEYS": {"keys": [{"key": "a_key"}]},
        "UPDATE_KEYS": {"index": "1-5"},
    }

    config = configsuite.ConfigSuite(invalid_missing_required_keyword, schema)
    assert not config.valid

    invalid_negative_index = {
        "CALCULATE_KEYS": {"keys": [{"key": "first_key"}, {"key": "second_key"}]},
        "UPDATE_KEYS": {"keys": [{"index": [-1, 2, 3], "key": "first_key"}]},
    }

    schema = job_config.build_schema()
    config = configsuite.ConfigSuite(invalid_negative_index, schema)
    assert not config.valid


def test_get_scaling_factor():
    new_event = namedtuple("named_dict", ["keys", "threshold"])
    event = new_event(["one_random_key"], 0.95)
    np.random.seed(123)
    input_matrix = np.random.rand(10, 10)

    matrix = scaled_matrix.DataMatrix(pd.DataFrame(data=input_matrix))

    assert matrix.get_scaling_factor(event) == np.sqrt(10 / 4.0)


@pytest.mark.parametrize(
    "threshold,expected_result", [(0.0, 1), (0.83, 2), (0.90, 3), (0.95, 4), (0.99, 6)]
)
def test_get_nr_primary_components(threshold, expected_result):
    np.random.seed(123)
    input_matrix = np.random.rand(10, 10)
    matrix = scaled_matrix.DataMatrix
    assert matrix._get_nr_primary_components(input_matrix, threshold) == expected_result


def test_std_normalization():
    input_matrix = pd.DataFrame(np.ones((3, 3)))
    input_matrix.loc["OBS"] = np.ones(3)
    input_matrix.loc["STD"] = np.ones(3) * 0.1
    expected_matrix = [[10.0, 10.0, 10.0], [10.0, 10.0, 10.0], [10.0, 10.0, 10.0]]
    matrix = scaled_matrix.DataMatrix(pd.concat({"A_KEY": input_matrix}, axis=1))
    result = matrix.std_normalization(["A_KEY"])
    assert (result.loc[[0, 1, 2]].values == expected_matrix).all()


def test_filter_on_column_index():
    matrix = np.random.rand(10, 10)

    index_lists = [[0, 1], [1, 2, 3], [1, 2, 3, 4, 5]]
    for index_list in index_lists:
        result = measured.MeasuredData._filter_on_column_index(
            pd.DataFrame(matrix), index_list
        )
        assert result.shape == (10, len(index_list))

    with pytest.raises(IndexError):
        measured.MeasuredData._filter_on_column_index(pd.DataFrame(matrix), [11])


@pytest.mark.parametrize(
    "matching_keys,entry,expected_result",
    [
        (["a_key"], {"key": "a_*"}, [{"key": "a_key"}]),
        (["a_key", "b_key"], {"key": "*key"}, [{"key": "a_key"}, {"key": "b_key"}]),
        (
            ["a_key"],
            {"key": "a_*", "index": [1, 2]},
            [{"key": "a_key", "index": [1, 2]}],
        ),
    ],
)
def test_wildcard_to_dict_list(matching_keys, entry, expected_result):
    assert job._wildcard_to_dict_list(matching_keys, entry) == expected_result


def test_find_and_expand_wildcards():
    expected_dict = {
        "ANOTHER_KEY": "something",
        "CALCULATE_KEYS": {
            "keys": [
                {"key": "WOPR_OP1_108"},
                {"key": "WOPR_OP1_144"},
                {"key": "WOPR_OP1_190"},
                {"key": "WOPR_OP1_9"},
                {"key": "WOPR_OP1_36"},
                {"key": "WOPR_OP1_72"},
                {"key": "FOPR"},
            ]
        },
        "UPDATE_KEYS": {
            "keys": [
                {"key": "WOPR_OP1_108"},
                {"key": "WOPR_OP1_144"},
                {"key": "WOPR_OP1_190"},
                {"key": "FOPR"},
            ]
        },
    }

    user_config = {
        "ANOTHER_KEY": "something",
        "CALCULATE_KEYS": {"keys": [{"key": "WOPR_*"}, {"key": "FOPR"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WOPR_OP1_1*"}, {"key": "FOPR"}]},
    }

    observation_list = [
        "WOPR_OP1_108",
        "WOPR_OP1_144",
        "WOPR_OP1_190",
        "WOPR_OP1_9",
        "WOPR_OP1_36",
        "WOPR_OP1_72",
        "FOPR",
    ]

    result_dict = job._find_and_expand_wildcards(observation_list, user_config)

    assert result_dict == expected_dict


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_add_observation_vectors():

    valid_config_data = {"UPDATE_KEYS": {"keys": [{"key": "WOPR_OP1_108"}]}}

    schema = job_config.build_schema()
    config = configsuite.ConfigSuite(valid_config_data, schema)

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil_field")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)

    obs = ert.getObservations()

    new_events = job._create_active_lists(obs, config.snapshot.UPDATE_KEYS.keys)

    keys = [event.key for event in new_events]

    assert "WOPR_OP1_108" in keys
    assert "WOPR_OP1_144" not in keys


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_validate_failed_realizations():
    """
    Config has several failed realisations
    """
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "custom_kw")
    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("mini_fail_config")
    ert = EnKFMain(res_config)
    observations = ert.getObservations()

    result = job.keys_with_data(
        observations,
        ["GEN_PERLIN_1"],
        ert.getEnsembleSize(),
        ert.getEnkfFsManager().getCurrentFileSystem(),
    )
    assert result == ["GEN_PERLIN_1"]


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_validate_no_realizations():
    """
    Ensamble has not run
    """
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "poly_normal")
    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("poly.ert")
    ert = EnKFMain(res_config)
    observations = ert.getObservations()

    result = job.keys_with_data(
        observations,
        ["POLY_OBS"],
        ert.getEnsembleSize(),
        ert.getEnkfFsManager().getCurrentFileSystem(),
    )
    assert result == []



#!/usr/bin/env python
import os
from setuptools import setup


def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


job_files = package_files("semeio/jobs/configs") + package_files("semeio/jobs/scripts")

setup(
    name="semeio",
    use_scm_version={"write_to": "semeio/version.py"},
    author="Software Innovation Bergen, Equinor ASA",
    author_email="fg_gpl@statoil.com",
    url="https://github.com/equinor/semeio",
    description="Jobs and workflow jobs for Ert.",
    packages=[
        "semeio",
        "semeio.hook_implementations",
        "semeio.jobs.correlated_observations_scaling",
    ],
    entry_points={"ert": ["semeio_jobs = semeio.hook_implementations.jobs",]},
    license="GPL-3.0",
    platforms="any",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=["configsuite", "numpy", "pandas", "six", "scipy",],
    setup_requires=["pytest-runner", "setuptools_scm"],
    tests_require=["pytest", "mock", 'black; python_version>="3.6"'],
    test_suite="tests",
    package_data={"": job_files},
    include_package_data=True,
)



import sys
import os

import pytest


@pytest.mark.skipif(sys.version_info < (3, 6), reason="requires Python3")
def test_code_style():
    from pathlib import Path
    import black
    from click.testing import CliRunner

    root = str(Path(__file__).parent.parent)

    runner = CliRunner()
    resp = runner.invoke(
        black.main,
        [
            "--check",
            os.path.join(root, "tests"),
            os.path.join(root, "semeio"),
            os.path.join(root, "setup.py"),
            "--exclude",
            "semeio/version.py",
        ],
    )

    assert (
        resp.exit_code == 0
    ), "Black would still reformat one or more files:\n{}".format(resp.output)



from ert_shared.plugins.plugin_manager import hook_implementation
from ert_shared.plugins.plugin_response import plugin_response


@hook_implementation
@plugin_response(plugin_name="semeio")
def installable_jobs():
    return {}


@hook_implementation
@plugin_response(plugin_name="semeio")
def installable_workflow_jobs():
    return {}
